#include<stdio.h>
#include<mpi.h>
#if THREADED == 1
  #include<pthread.h>
#endif
#include<string.h>
/* there are more library specific includes later! */

typedef enum {NOskel, Fskel, SFskel, Pskel, Bskel} skeltag_t;

typedef enum {SPLOCAL, SPGLOBAL} spread_t;

typedef struct {
  MPI_Comm parentcomm;
  int      parentcommsize;
  int      parentcommrank;

  MPI_Comm tofarmercomm;    /* for sending messages to the farmer     */
  MPI_Comm fromfarmercomm;  /* for receiving messages from the farmer */

  spread_t inspr;       /* input spread  */
  MPI_Datatype inty;    /* input type    */
  int insz;             /* input size    */
  spread_t outspr;      /* output spread */
  MPI_Datatype outty;   /* output type   */
  int outsz;            /* output size   */

  int w;                /* number of workers */
  int *wsz;             /* (length w) number of processes in each worker */
  int **wids;           /* (length w of length wsz[i]) parentcomm ids of processes in each worker */

  char     **outbuff;         /* the worker-wise output buffer */
  int      *outblens;         /* local lengths of the contributions to output buffer   */
  int      *outbtags;         /* tags indicating the item's originating input task     */
  int      outbcount;         /* accumulates the length of the output buffer           */
  int      currenttag;        /* where did any items which are Given originate         */

  /* this bit relates to the lazy, "one farmer" approach only                  */
  int      allinmul;         /* number of items in the inbuff, once gathered          */
  char     **inbuff;         /* the input buffer, active on upper P0 only             */
  int      *inblens;         /* lengths of the inmult contributions to the buffer     */
  int      inbnext;          /* marker for running through buffer                     */

} df_t;

typedef struct {
  MPI_Comm parentcomm;
  int      parentcommsize;
  int      parentcommrank;

  MPI_Comm tofarmercomm;    /* for sending messages to the farmer     */
  MPI_Comm fromfarmercomm;  /* for receiving messages from the farmer */

  spread_t inspr;       /* input spread  */
  MPI_Datatype inty;    /* input type    */
  int insz;             /* input size    */
  spread_t outspr;      /* output spread */
  MPI_Datatype outty;   /* output type   */
  int outsz;            /* output size   */

  char     **outbuff;         /* the worker-wise output buffer */
  int      *outblens;         /* local lengths of the contributions to output buffer   */
  int      *outbtags;         /* tags indicating the item's originating input task     */
  int      outbcount;         /* accumulates the length of the output buffer           */
  int      currenttag;        /* where did any items which are Given originate         */

  /* this bit relates to the lazy, "one farmer" approach only                  */
  int      allinmul;         /* number of items in the inbuff, once gathered          */
  char     **inbuff;         /* the input buffer, active on upper P0 only             */
  int      *inblens;         /* lengths of the inmult contributions to the buffer     */
  int      inbnext;          /* marker for running through buffer                     */

} sf_t;

/* structural information about a pipeline which is required by other functions      */
/* eg Give and Take, and which must therefore be stored in the run time environment  */
/* Note that some pipeline parameters (eg stages) are not stored here since they are */
/* only ever referenced from within the body of Pipeline() itself.                   */
typedef struct {
  int      ns;
  MPI_Comm parentcomm;
  int      parentcommsize;
  int      parentcommrank;
  spread_t *spr;  /* array of spreads */
  MPI_Datatype *ty;   /* array of types */
  int      inmul;
  int      shutdown;
  int      mystage;

  MPI_Comm allcomm1;       /* a duplicate of parentcomm, for group2group */
  MPI_Comm allcomm2;       /* a duplicate of parentcomm, for group2group */
  MPI_Comm thisstagecomm;  /* a duplicate of comm,       for group2group */
  MPI_Comm thisstagerootcomm;   /* a comm containing only the root of the stage */
                           /* required as a (probably unused) parameter to g2g */ 
  int      *allranks;      /* all the ranks in parentcomm, in order */

  int      *allcols;       /* array of stage ids, as selected by all pipeline members */

  int      *nextstage;     /* array of proc ids, in ascending order (so nextstage[0] is the id of */ 
                           /* the root of the next stage with respect to the parent comm          */
  int      nextstagesize;  /* active length of above */

  int      *prevstage;     /* as above for the preceding stage */

  int      prevstagesize;  

  int      *firststage;    
  int      firststagesize;  

  int      *laststage;     /* as above for the last  stage */
  int      laststagesize; 

  int      *thisstage;     /* as above for the local stage */
  int      thisstagesize;  

  int      *firststageroot; /* as above for specical group containing only first stage root */
                            /* no need for a stagesize for this  - it is 1                  */

  int      *laststageroot;  /* as above for specical group containing only last stage root */
                            /* no need for a stagesize for this  - it is 1                 */

  int      *itemsz;        /* ns+1 sizes of raw stage to stage items */

  char     **buff;         /* the real stage 0 buffer, active in all st 0 processes */
                           /* for SPGLOBAL data or the root only for SPLOCAL data   */
  int      *blens;         /* local lengths of the mult contributions to the buffer */
                           /* or root only lengths for SP_LOCAL data                */
  int      bnext;          /* marker for running through buffer                     */

  char     **poutbuff;         /* the real last stage buffer, active in all last st processes */
  int      *poutblens;         /* local lengths of the poutbnext contributions to this buffer */
  int      poutbnext;          /* marker for running through buffer, counts size              */
} p_t;


/* structural information about a butterfly */
typedef struct {
  int      nd;
  int      mypartnerrank;     /* partner at the currently active level */
  MPI_Datatype ty;          /* type for transfers */
  int      itemsz;          /* size of raw item   */
} b_t;

typedef struct {
  skeltag_t skel;
  MPI_Comm  mycomm;
  int       mycommsize;
  int       myrank;
  union {
    df_t f;
    sf_t sf;
    p_t  p;
    b_t  b;
  } context;
} skel_t;

typedef struct {
  void *data;
  int   len;
} eSkel_atom_t;

#define MAXDEPTH 6
#define MAXOUTMUL 100000      
#define NOSHUTDOWNTAG 1      /* tag for Give-Take to indicate more data to come */
#define FARMTAKETAG 2        /* tag with which a worker requests a task of a farmer */
#define SHUTDOWNTAG 0        /* tag for Give-Take to indicate no more data to come */
#define SHUTDOWNLEN -999     /* tag for Give-Take to indicate no more data to come */
skel_t *FrStack[MAXDEPTH];

#define SENDER   0
#define RECEIVER 1

#define FALSE 0 
#define TRUE  1

int somewhere;               /* because I need a safe non-null address to return from SPLOCAL takes */
int mpiintsize;
int zero = 0;
int Top;


/**************************************************************************************************/

void SkelLibInit (void);

int myrank (void);

int mycommsize (void);

MPI_Comm mycomm (void);

void  Give (void *out, int len);

void *Take (int *items);

void *Exchange (int *inlen, void *out, int outlen);

void Pipeline (int ns, void (*stages[])(),
               int col, spread_t spr[], MPI_Datatype ty[], void *in, int inlen, int inmul,
               void *out, int outlen, int *outmul, int outbuffsz, MPI_Comm comm);

void Pipelinev (int ns, void (*stages[])(),
                int col, spread_t spr[], MPI_Datatype ty[], void *in, int *inlen, int inmul,
                void *out, int *outlen, int *outmul, MPI_Comm comm);

void Pipeline1for1 (int ns, eSkel_atom_t *(*stages[]) (eSkel_atom_t *),
                    int col, spread_t spr[], MPI_Datatype ty[],
                    void *in, int inlen, int inmul, void *out, int outlen, int *outmul,
                    int outbuffsz, MPI_Comm comm);

void Pipeline1for1v (int ns, eSkel_atom_t *(*stages[])(eSkel_atom_t *),
                     int col, spread_t spr[], MPI_Datatype ty[],
                     void *in, int *inlen, int inmul,
		     void *out, int *outlen, int *outmul, MPI_Comm comm);

void PipeGive (void *out, int len);

void *PipeTake (int *items);





#if THREADED == 1
void Farm (int nw, void worker(void), int col, 
            void *in,  int inlen, int inmul, spread_t inspr, MPI_Datatype inty, 
            void *out, int outlen, int *outmul, spread_t outspr, MPI_Datatype outty, 
            int outbuffsz, MPI_Comm comm);

void Farmv (int nw, void worker(void), int col, 
            void *in,  int *inlen, int inmul, spread_t inspr, MPI_Datatype inty, 
	    void *out, int *outlen, int *outmul, spread_t outspr, MPI_Datatype outty,  
            MPI_Comm comm);

void Farm1for1 (int nw, eSkel_atom_t * worker (eSkel_atom_t *), int col,
                void *in,  int inlen, int inmul, spread_t inspr, MPI_Datatype inty, 
		void *out, int outlen, int *outmul, spread_t outspr, MPI_Datatype outty,  
                int outbuffsz, MPI_Comm comm);

void Farm1for1v (int nw, eSkel_atom_t * worker (eSkel_atom_t *), int col, 
                 void *in,  int *inlen, int inmul, spread_t inspr, MPI_Datatype inty, 
      	         void *out, int *outlen, int *outmul, spread_t outspr, MPI_Datatype outty, 
                 MPI_Comm comm);
#endif
void FarmGive (void *out, int len);

void *FarmTake (int *items);




void SimpleFarm (void worker(void),
                 void *in,  int inlen, int inmul, spread_t inspr, MPI_Datatype inty, 
		 void *out, int outlen, int *outmul, spread_t outspr, 
                 MPI_Datatype outty, int outbuffsz, MPI_Comm comm);

void SimpleFarmv (void worker(void), void *in,  int *inlen, int inmul, spread_t inspr, 
                  MPI_Datatype inty, 
                  void *out, int *outlen, int *outmul, spread_t outspr, MPI_Datatype outty, 
                  MPI_Comm comm);

void SimpleFarm1for1 (eSkel_atom_t *worker (eSkel_atom_t *),
                      void *in,  int inlen, int inmul, spread_t inspr, MPI_Datatype inty,
                      void *out, int outlen, int *outmul, spread_t outspr, MPI_Datatype outty, 
                      int outbuffsz, MPI_Comm comm);

void SimpleFarm1for1v (eSkel_atom_t * worker (eSkel_atom_t *), void *in,  int *inlen, int inmul,
                       spread_t inspr, MPI_Datatype inty, void *out, int *outlen, int *outmul, 
                       spread_t outspr, MPI_Datatype outty, MPI_Comm comm);

void SimpleFarmGive (void *out, int len);

void *SimpleFarmTake (int *items);



void Butterfly (int nd, void level(), MPI_Datatype ty, MPI_Comm comm);

void ButterflyGive (void *out, int len);

void *ButterflyTake (int *items);

void *ButterflyExchange (int *inlen, void *out, int outlen);

/******************************************************************************************************/

#include "eSkel_generic.c"
#include "eSkel_utils.c"

#include "butterflysupport.c"
#include "butterfly.c"

#include "farmsupport.c"
#include "farmsorter.c"
#include "simplefarmsupport.c"
#include "simplefarmv.c" 
#include "simplefarm.c"
#include "simplefarm1for1v.c"
#include "simplefarm1for1.c"

#include "pipelinesupport.c"

#if THREADED == 1
  #include "pipelinev.c"
#elif THREADED == 0 
  #include "pipelinev-no-threads.c"
#endif

#include "pipeline.c"

#if THREADED == 1
  #include "pipeline1for1v.c"
#elif THREADED == 0 
  #include "pipeline1for1v-no-threads.c"
#endif

#include "pipeline1for1.c"

#if THREADED == 1
#include "farmv.c"
#include "farm.c"
#include "farm1for1.c"
#include "farm1for1v.c"
#endif
